class Array
  def deep_copy
    map{|x|
      if (x.class == Array)
        x.deep_copy
      else
        x
      end
    }
  end
end

class Binformat
  @@format = {}
  @@checked_constants = {}
  @@constants = {}

  def Binformat.method_missing(s, *a)
    (defined?(@@val) && @@val[s]) || super(s, *a)
  end

  def parse_elem(f, elem)
    type, sym, cntexp, *args = *elem
    cntcls = cntexp.class
    cnt = if cntcls == Symbol
            @val[cntexp]
          elsif cntcls == Proc && type != :repeat
            @@val = @val
            r = cntexp[]
            @@val = nil
            r
          else
            cntexp
          end

    case type
    when :seek
      f.seek(cnt)
      nil
    when :endian
      if (sym != :machine && sym != :big && sym != :little)
        raise sym.to_s + ": unknown endian"
      end
      @endian = sym
      nil
    when :byte
      d = f.read(cnt)
      d = d[0] if cntcls == Fixnum && cnt == 1
      @val[sym] = d
      [sym, d]
    when :short
      d = f.read(2*cnt).unpack(endian_string(type, @endian))
      d = d[0] if cntcls == Fixnum && cnt == 1
      @val[sym] = d
      [sym, d]
    when :int
      d = f.read(4*cnt).unpack(endian_string(type, @endian))
      d = d[0] if cntcls == Fixnum && cnt == 1
      @val[sym] = d
      [sym, d]
    when :struct
      r = []
      cnt.times {
        c = args[0].send(:new)
        d = c.parse(f)
        r << d
      }
      @val[sym] = r
      [sym, r]
    when :repeat
      if args[0].class == Class
        cls = args[0]
        done = cnt
        r = []
        c = cls.send(:new)
        while true
          d = c.parse(f)
          r << d
          break if done[d]
        end
        [sym, r]
      else
        raise 'We only support struct in repeat'
      end
    when :some
      prc = sym
      @@running = [self, f]
      @@val = @val
      d = prc[]
      raise "strange proc???" if !d
      @@running = nil
      @@val = nil
      d
    end
  end

  def dump_elem(l, elem)
    type, sym, cntexp, *args = *elem
    cntcls = cntexp.class
    cnt = if cntcls == Symbol
            @val[cntexp]
          elsif cntcls == Proc && type != :repeat
            @@val = @val
            r = cntexp[]
            @@val = nil
            r
          else
            cntexp
          end

    case type
    when :endian
      if (sym != :machine && sym != :big && sym != :little)
        raise sym.to_s + ": unknown endian"
      end
      @endian = sym
      nil
    when :byte
      s, v = l.shift
      assert(s == sym)
      @val[sym] = v
      if v.class != String
        x = ''
        x << v
        v = x
      end
      assert(v.size == cnt)
      v
    when :short
      s, v = l.shift
      assert(s == sym)
      @val[sym] = v
      v = [v] if v.class != Array
      assert(v.size == cnt)
      v.pack(endian_string(type, @endian))
    when :int
      s, v = l.shift
      assert(s == sym)
      @val[sym] = v
      v = [v] if v.class != Array
      assert(v.size == cnt)
      v.pack(endian_string(type, @endian))
    when :struct
      s, v = l.shift
      assert(s == sym)
#      assert(v.cnt == cnt)
      r = ''
#      puts "#{sym} #{v.inspect} (dump)"
      @val[sym] = v.deep_copy
      cnt.times {|i|
        c = args[0].send(:new)
        d = c.dump(v[i])
        r << d
      }
      r
    when :repeat
      if args[0].class == Class
        cls = args[0]
        done = cnt
        r = ''
        c = cls.send(:new)
        s, v = l.shift
        assert(s == sym)
        @val[sym] = v.deep_copy
        v.each do |d|
          r << c.dump(d)
        end
        r
      else
        raise 'We only support struct in repeat'
      end
    when :some
      prc = sym
      @@dumping = [self, l]
      @@val = @val
      d = prc[]
      raise "strange proc???" if !d
      @@dumping = nil
      @@val = nil
      d
    end
  end

  def parse(f)
    ret = []
    @val = {}
    @endian = :machine
    if @@format[self.class]
      @@format[self.class].each do |elem|
        v = parse_elem(f, elem)
        ret << v if v
        #p v
      end
    end
    ret
  end

  def dump(l)
    ret = ''
    @val = {}
    @endian = :machine
    if @@format[self.class]
      @@format[self.class].each do |elem|
        v = dump_elem(l, elem)
        ret << v if v
        #p v
      end
    end
    ret
  end

  def Binformat.byte(sym, cnt=1)
    add_format [:byte, sym, cnt]
  end

  def Binformat.short(sym, cnt=1)
    add_format [:short, sym, cnt]
  end

  def Binformat.int(sym, cnt=1)
    add_format [:int, sym, cnt]
  end
  def Binformat.addr(sym, cnt=1)
    int(sym, cnt)
  end

  def Binformat.struct(cls, sym, cnt=1)
    add_format [:struct, sym, cnt, cls]
  end

  def Binformat.repeat(cls, sym, done)
    add_format [:repeat, sym, done, cls]
  end

  def Binformat.endian(sym)
    add_format [:endian, sym]
  end

  def Binformat.some(&proc)
    add_format [:some, proc]
  end

  def Binformat.seek(cnt)
    add_format [:seek, nil, cnt]
  end

private
  def endian_string(type, endian)
    if (endian == :machine)
      s = {
        :short => 'S*',
        :int => 'L*',
      }[type]
    else
      s = {
        :big => 'n*',
        :little => 'v*',
      }[endian]
      if (type == :int)
        s.upcase!
      end
    end
    s
  end

  def Binformat.check_constants(sym)
    if (!@@constants.key?(self))
      @@constants[self] = {}
    end
    if (!@@checked_constants.key?(self))
      @@checked_constants[self] = []
    end
    c = constants - @@checked_constants[self]
    return if c.empty?
    @@checked_constants[self] += c
    @@constants[self][sym] = c
  end

  def Binformat.add_format(a)
    if (defined?(@@running) && @@running)
      @@running[0].parse_elem(@@running[1], a)
    elsif (defined?(@@dumping) && @@dumping)
      @@dumping[0].dump_elem(@@dumping[1], a)
    else
      check_constants(a[1])

      if (!@@format.key?(self))
        @@format[self] = []
      end
      @@format[self] << a
    end
  end

  def assert(bt)
#    raise "assert" if !bt
  end

end

def show_binary(b, t=0)
  if b.class == Array
    if (b[0].class == Symbol)
      ' '*t+"["+b.map{|x|show_binary(x,t+1)}*', '+"]"
    else
      s=' '*t+"[\n"
      b.each do |x|
        s+=show_binary(x,t+1)+",\n"
      end
      s+=' '*t+"]\n"
    end
  elsif b.class == Symbol
    b.inspect
  elsif b.class == Fixnum
    b.inspect
  elsif b.class == String
    b.inspect
  end
end
