#!/usr/bin/env ruby

require 'optparse'
require 'yaml'
require 'pp'

TEXT_TYPES = {
  'p' => [proc{|v|v.pretty_inspect},
          proc{|v|eval(v)}],
  'yml' => [proc{|v|YAML.dump(v)},
            proc{|v|YAML.load(v)}],
}

BINARY_EXT_ALIASES = {
  'o' => 'elf',
  'out' => 'elf',
}

text_type = nil
bin_type = nil
reverse = false

input_file = nil
output_file = nil

OptionParser.new("Usage: #$0 [options] input [output]") do |opt|
  opt.on('-y', 'use YAML as text format'){|v|text_type = 'yml'}
  opt.on('-t TYPE', 'type of binary'){|v|bin_type = v}
  opt.on('-r', 'reverse: generate binary from text dump'){|v|reverse = true}

  opt.parse!(ARGV)

  input_file, output_file = *ARGV

  if !input_file
    print opt.help
    exit 1
  end
end

input_ext = File.extname(input_file)[1..-1]

if bin_type
  if !File.exist?("#{bin_type}.rb")
    raise "No driver for #{bin_type}"
  end
else
  if BINARY_EXT_ALIASES[input_ext]
    bin_type = BINARY_EXT_ALIASES[input_ext]
  elsif File.exist?("#{input_ext}.rb")
    bin_type = input_ext
  else
    if output_file
      output_ext = File.extname(output_file)[1..-1]
      if File.exist?("#{output_ext}.rb")
        bin_type = output_ext
        reverse = true
      end
    end
    if !bin_type
      raise "No driver for #{input_file}"
    end
  end
end

if !text_type
  if TEXT_TYPES[input_ext]
    reverse = true
  end

  text_file = reverse ? input_file : output_file
  if text_file && File.extname(text_file) == '.yml'
    text_type = 'yml'
  else
    text_type = 'p'
  end
end

require bin_type
bin_parser = $BINFORMAT_PROVIDE.send(:new)

output = $>
if reverse && !output_file && output.tty?
  output_file = "out.#{bin_type}"
  $>.puts "Attempt to output binary to tty, use #{output_file} instead."
end
if output_file
  output = File.open(output_file, 'w')
end

if reverse
  output.print bin_parser.dump(TEXT_TYPES[text_type][1][File.read(input_file)])
else
  File.open(input_file) do |input|
    output.print TEXT_TYPES[text_type][0][bin_parser.parse(input)]
  end
end

if output_file
  output.close
end
