require 'binformat'

class GIFFileHeader < Binformat
  byte :signature, 3
  byte :version, 3
end

class GIFRGB < Binformat
  byte :r
  byte :g
  byte :b
end

class GIFLogicalScreenDescriptor < Binformat
  short :width
  short :height
  byte :packed
  byte :bgcolor
  byte :aspect_ratio

  struct GIFRGB, :globalColorTable, proc{
    (packed >> 7) == 0 ? 0 : (1 << ((packed & 0b111) + 1))
  }
end

class GIFDataSubBlock < Binformat
  byte :size
  byte :data, :size
end

class GIFImageDescriptor < Binformat
  short :left
  short :top
  short :width
  short :height
  byte :packed

  struct GIFRGB, :localColorTable, proc{
    (packed >> 7) == 0 ? 0 : (1 << ((packed & 0b111) + 1))
  }

  byte :lzwMinimumCodeSize
  repeat GIFDataSubBlock, :imageData, proc{|data|
    data.assoc(:size)[1] == 0
  }
end

class GIFGraphicControlExtension < Binformat
  byte :blockSize
  byte :packed
  short :delayTime
  byte :transparentColorIndex
  byte :blockTerminator
end

class GIFCommentExtension < Binformat
  repeat GIFDataSubBlock, :commentData, proc{|data|
    data.assoc(:size)[1] == 0
  }
end

class GIFApplicationExtension < Binformat
  byte :blockSize
  byte :appIdentifier, 8
  byte :appAuthenticationCode, 3
  repeat GIFDataSubBlock, :appData, proc{|data|
    data.assoc(:size)[1] == 0
  }
end

class GIFPlainTextExtension < Binformat
  byte :blockSize
  short :gridLeft
  short :gridTop
  short :gridWidth
  short :gridHeight
  byte :charCellWidth
  byte :charCellHeight
  byte :fgColor
  byte :bgColor
  repeat GIFDataSubBlock, :textData, proc{|data|
    data.assoc(:size)[1] == 0
  }
end

class GIFExtension < Binformat
  byte :label
  some {
    #p label
    case label
    when 0xf9
      struct GIFGraphicControlExtension, :graphicControlExtension
    when 0xfe
      struct GIFCommentExtension, :commentExtension
    when 0xff
      struct GIFApplicationExtension, :applicationExtension
    when 0x01
      struct GIFPlainTextExtension, :plainTextExtension
    else
      raise "Unknown exetension label: #{label}"
    end
  }
end

class GIFTrailer < Binformat
end

class GIFData < Binformat
  byte :tag
  some {
    #p tag
    case tag
    when 0x2c
      struct GIFImageDescriptor, :imageDescriptor
    when 0x21
      struct GIFExtension, :extension
    when 0x3b
      struct GIFTrailer, :trailer
    else
      raise "Unknown separator: #{tag}"
    end
  }
end

class GIF < Binformat
  struct GIFFileHeader, :fileHdr
  struct GIFLogicalScreenDescriptor, :logicalScreenDescriptor

  repeat GIFData, :data, proc{|data|
    data.assoc(:tag)[1] == 0x3b
  }
end

$BINFORMAT_PROVIDE = GIF

if $0 == __FILE__
  require 'yaml'

  f = GIF.new
  #r = f.parse(ARGF)
  #puts r.to_yaml
  print f.dump(YAML.load(File.read(ARGV[0])))
end
